import numpy
import scipy.linalg


tauBzero = 1.520e-12 * 1. / 6.582119e-16 / 1e-9
tauBplus = 1.638e-12 * 1. / 6.582119e-16 / 1e-9


class BaBarBToRho:
    """
    Reference:
    """
    def __init__(self):
        babar_bins = [[0, 8], [8, 16], [16, 20.3]]
        babar_central = numpy.array([0.564, 0.912, 0.268]) * 1e-4
        babar_central = numpy.array([x / (b[1] - b[0]) for x, b in zip(babar_central, babar_bins)])

        babar_covariance = numpy.array([
            [4.298, 2.188, 1.137],
            [0, 3.381, 1.034],
            [0, 0, 2.086],
        ])
        babar_covariance = (babar_covariance + babar_covariance.T - numpy.diag(babar_covariance.diagonal())) * 1e-12

        self.bins = babar_bins
        self.central = babar_central / (tauBplus / 2)
        self.covariance = babar_covariance / (tauBplus / 2) ** 2


class BelleBToRho:
    """
    Reference:
    """
    def __init__(self):
        belle_neutral_bins = [[i, i+4] for i in range(0, 24, 4)]  # Maybe 24 = 20.3?
        belle_neutral_central = numpy.array([37.3, 71.8, 80.6, 72.3, 62.6, 1.7]) * 1e-6
        belle_neutral_central = numpy.array([x / (b[1] - b[0]) for x, b in zip(belle_neutral_central, belle_neutral_bins)])

        belle_neutral_error = numpy.array([10.6, 11.6, 12.3, 12.5, 11.5, 7.9]) * 1e-6
        belle_neutral_error = numpy.array([x / (b[1] - b[0]) for x, b in zip(belle_neutral_error, belle_neutral_bins)])

        # belle_neutral_covariance = numpy.diag(belle_neutral_error ** 2)  # correlations below

        belle_neutral_correlation = numpy.array([
            [100, -14.1, 2.4, 0.9, 1.1, 1.6],
            [0, 100, -11.2, 3.0, 2.4, 2.5],
            [0, 0, 100, -7.5, 5.0, 5.8],
            [0, 0, 0, 100, -8.8, 6.1],
            [0, 0, 0, 0, 100, -10.2],
            [0, 0, 0, 0, 0, 100],
        ]) / 100
        belle_neutral_correlation = (belle_neutral_correlation + belle_neutral_correlation.T) - numpy.diag(
            belle_neutral_correlation.diagonal())
        belle_neutral_covariance_statistical = numpy.diag(belle_neutral_error) @ belle_neutral_correlation @ numpy.diag(
            belle_neutral_error)

        tauBzero = 1.520e-12 * 1. / 6.582119e-16 / 1e-9
        tauBplus = 1.638e-12 * 1. / 6.582119e-16 / 1e-9

        belle_neutral_central *= (tauBplus / 2 / tauBzero)
        belle_neutral_covariance_statistical *= (tauBplus / 2 / tauBzero)

        belle_charged_bins = [[i, i + 2] for i in range(0, 22, 2)]
        belle_charged_central = numpy.array([12.4, 15.3, 17.9, 21.7, 14.3, 19.9, 22.4, 21.7, 17.5, 14.3, 5.9]) * 1e-6
        belle_charged_central = numpy.array([x / (b[1] - b[0]) for x, b in zip(belle_charged_central, belle_charged_bins)])

        belle_charged_error = numpy.array([3.4, 3.1, 3.4, 3.4, 2.9, 3.4, 3.4, 3.3, 3.2, 2.8, 2.6]) * 1e-6
        belle_charged_error = numpy.array([x / (b[1] - b[0]) for x, b in zip(belle_charged_error, belle_charged_bins)])

        # belle_charged_covariance = numpy.diag(belle_charged_error ** 2)  # correlations below

        belle_charged_correlation = numpy.array([
            [100, -12.3, 0.8, 0.0, 0.2, 0.2, 0.2, 0.3, 0.3, 0.3, 0.3],
            [0, 100, -10.9, 0.8, 0.5, 0.5, 0.6, 0.7, 0.9, 0.8, 0.8],
            [0, 0, 100, -9.7, 0.9, 0.9, 0.8, 0.9, 1.3, 1.1, 1.1],
            [0, 0, 0, 100, -8.8, 1.4, 1.2, 1.5, 1.8, 1.4, 1.6],
            [0, 0, 0, 0, 100, -9.3, 1.5, 1.6, 2.1, 1.7, 1.8],
            [0, 0, 0, 0, 0, 100, -6.6, 1.9, 2.7, 2.3, 2.4],
            [0, 0, 0, 0, 0, 0, 100, -4.8, 1.8, 1.8, 2.4],
            [0, 0, 0, 0, 0, 0, 0, 100, -3.9, 2.3, 2.5],
            [0, 0, 0, 0, 0, 0, 0, 0, 100, -3.8, 2.7],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 100, -5.2],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 100],
        ]) / 100
        belle_charged_correlation = (belle_charged_correlation + belle_charged_correlation.T) - numpy.diag(
            belle_charged_correlation.diagonal())
        belle_charged_covariance_statistical = numpy.diag(belle_charged_error) @ belle_charged_correlation @ numpy.diag(
            belle_charged_error)

        belle_charged_central = belle_charged_central / (tauBplus / 2)
        belle_neutral_central = belle_neutral_central / (tauBplus / 2)

        belle_charged_covariance_statistical = belle_charged_covariance_statistical / (tauBplus / 2) ** 2
        belle_neutral_covariance_statistical = belle_neutral_covariance_statistical / (tauBplus / 2) ** 2

        belle_central = numpy.concatenate((belle_charged_central, belle_neutral_central))
        belle_covariance_statistical = scipy.linalg.block_diag(belle_charged_covariance_statistical,
                                                               belle_neutral_covariance_statistical)

        systematic_error_belle_charged_tracking = 0.7 * 0.01
        systematic_error_belle_charged_pi0 = 0.0 * 0.01
        systematic_error_belle_charged_pid = 1.0 * 0.01
        systematic_error_belle_charged_kid = 2.0 * 0.01
        systematic_error_belle_charged_qq = 0.7 * 0.01
        systematic_error_belle_charged_crossfeed = 2.4 * 0.01
        systematic_error_belle_charged_tagcal = 4.2 * 0.01
        systematic_error_belle_charged_ffshape = 1.3 * 0.01

        systematic_error_belle_charged_all = [
            systematic_error_belle_charged_tracking,
            systematic_error_belle_charged_pi0,
            systematic_error_belle_charged_pid,
            systematic_error_belle_charged_kid,
            systematic_error_belle_charged_qq,
            systematic_error_belle_charged_crossfeed,
            systematic_error_belle_charged_tagcal,
            systematic_error_belle_charged_ffshape,
        ]

        systematic_error_belle_neutral_tracking = 0.35 * 0.01
        systematic_error_belle_neutral_pi0 = 2.0 * 0.01
        systematic_error_belle_neutral_pid = 1.0 * 0.01
        systematic_error_belle_neutral_kid = 1.0 * 0.01
        systematic_error_belle_neutral_qq = 0.5 * 0.01
        systematic_error_belle_neutral_crossfeed = 5.0 * 0.01
        systematic_error_belle_neutral_tagcal = 4.5 * 0.01
        systematic_error_belle_neutral_ffshape = 1.17 * 0.01

        systematic_error_belle_neutral_all = [
            systematic_error_belle_neutral_tracking,
            systematic_error_belle_neutral_pi0,
            systematic_error_belle_neutral_pid,
            systematic_error_belle_neutral_kid,
            systematic_error_belle_neutral_qq,
            systematic_error_belle_neutral_crossfeed,
            systematic_error_belle_neutral_tagcal,
            systematic_error_belle_neutral_ffshape,
        ]

        systematic_error_belle_all = systematic_error_belle_charged_all + systematic_error_belle_neutral_all
        systematic_uncertainties_belle = [systematic * belle_central for systematic in systematic_error_belle_all]
        belle_covariance_systematic = sum([numpy.outer(s, s) for s in systematic_uncertainties_belle])

        belle_covariance = belle_covariance_statistical + belle_covariance_systematic

        self.charged_bins = belle_charged_bins
        self.charged_central = belle_charged_central
        self.charged_covariance = belle_covariance[:len(belle_charged_bins), :len(belle_charged_bins)]
        self.charged_covariance_statistical = belle_covariance_statistical[:len(belle_charged_bins), :len(belle_charged_bins)]
        self.charged_covariance_systematic = belle_covariance_systematic[:len(belle_charged_bins), :len(belle_charged_bins)]

        self.neutral_bins = belle_neutral_bins
        self.neutral_central = belle_neutral_central
        self.neutral_covariance = belle_covariance[len(belle_charged_bins):, len(belle_charged_bins):]
        self.neutral_covariance_statistical = belle_covariance_statistical[len(belle_charged_bins):, len(belle_charged_bins):]
        self.neutral_covariance_systematic = belle_covariance_systematic[len(belle_charged_bins):, len(belle_charged_bins):]

        self.bins = [*belle_charged_bins, *belle_neutral_bins]
        self.central = belle_central
        self.covariance = belle_covariance
        self.covariance_statistical = belle_covariance_statistical
        self.covariance_systematic = belle_covariance_systematic


class BaBarBToOmega:
    """
    Reference:
    """
    def __init__(self):
        babar_bins = [[0, 4], [4, 8], [8, 10], [10, 12], [12, 21]]
        babar_central = numpy.array([0.214, 0.200, 0.147, 0.169, 0.482]) * 1e-4
        babar_central = numpy.array([x / (b[1] - b[0]) for x, b in zip(babar_central, babar_bins)])
        babar_central = babar_central / (tauBplus / 2)

        babar_error = numpy.array([0.060, 0.037, 0.029, 0.031, 0.093]) * 1e-4
        babar_error = numpy.array([x / (b[1] - b[0]) for x, b in zip(babar_error, babar_bins)])
        babar_error = babar_error / (tauBplus / 2)

        babar_correlation = numpy.array([
            [1.000, 0.121, 0.061, 0.001, -0.011],
            [0.000, 1.000, 0.105, 0.094, 0.128],
            [0.000, 0.000, 1.000, 0.088, 0.121],
            [0.000, 0.000, 0.000, 1.000, 0.125],
            [0.000, 0.000, 0.000, 0.000, 1.000]
        ])

        babar_correlation = (babar_correlation + babar_correlation.T) - numpy.diag(babar_correlation.diagonal())
        babar_covariance_statistical = numpy.diag(babar_error) @ babar_correlation @ numpy.diag(babar_error)

        systematic_error_babar_tracking = numpy.array([3.9, 1.5, 2.8, 2.3, 1.1]) * 0.01
        systematic_error_babar_photon = numpy.array([2.0, 1.7, 3.3, 1.1, 0.6]) * 0.01
        systematic_error_babar_klong = numpy.array([4.8, 1.8, 2.5, 1.1, 1.4]) * 0.01
        systematic_error_babar_lid = numpy.array([1.6, 1.5, 1.5, 1.2, 1.2]) * 0.01
        systematic_error_babar_veto = numpy.array([1.7, 1.7, 1.7, 1.7, 1.7]) * 0.01

        systematic_error_babar_signalff = numpy.array([6.3, 1.5, 1.1, 2.9, 4.6]) * 0.01
        systematic_error_babar_bromega = numpy.array([0.8, 0.8, 0.8, 0.8, 0.8]) * 0.01
        systematic_error_babar_radcorr = numpy.array([0.4, 0.3, 0.2, 0.1, 0.2]) * 0.01

        systematic_error_babar_qqshape = numpy.array([2.6, 0.1, 0.4, 0.2, 0.3]) * 0.01
        systematic_error_babar_bbshape = numpy.array([2.0, 0.9, 1.8, 0.2, 0.1]) * 0.01
        systematic_error_babar_btoxc = numpy.array([0.2, 0.6, 0.3, 0.2, 0.2]) * 0.01
        systematic_error_babar_btoxu = numpy.array([0.3, 0.4, 0.4, 0.3, 0.5]) * 0.01

        systematic_error_babar_comb1 = numpy.array([0.6, 0.5, 0.4, 1.1, 3.7]) * 0.01
        systematic_error_babar_comb2 = numpy.array([4.2, 1.0, 0.9, 0.9, 2.0]) * 0.01
        systematic_error_babar_comb3 = numpy.array([1.7, 2.2, 2.7, 2.7, 3.5]) * 0.01

        systematic_error_babar_totbb = numpy.array([1.1, 1.1, 1.1, 1.1, 1.1]) * 0.01
        systematic_error_babar_fpm = numpy.array([1.2, 1.2, 1.2, 1.2, 1.2]) * 0.01

        systematic_error_babar_all = [
            systematic_error_babar_tracking,
            systematic_error_babar_photon,
            systematic_error_babar_klong,
            systematic_error_babar_lid,
            systematic_error_babar_veto,
            systematic_error_babar_signalff,
            systematic_error_babar_bromega,
            systematic_error_babar_radcorr,
            systematic_error_babar_qqshape,
            systematic_error_babar_bbshape,
            systematic_error_babar_btoxc,
            systematic_error_babar_btoxu,
            systematic_error_babar_comb1,
            systematic_error_babar_comb2,
            systematic_error_babar_comb3,
            systematic_error_babar_totbb,
            systematic_error_babar_fpm
        ]

        systematic_uncertainties_babar = [systematic * babar_central for systematic in systematic_error_babar_all]
        babar_covariance_systematic = sum([numpy.outer(s, s) for s in systematic_uncertainties_babar])

        babar_covariance = babar_covariance_statistical + babar_covariance_systematic

        self.bins = babar_bins
        self.central = babar_central
        self.covariance = babar_covariance
        self.covariance_statistical = babar_covariance_statistical
        self.coviance_systematic = babar_covariance_systematic


class BelleBToOmega:
    """
    Reference:
    """
    def __init__(self):
        belle_bins = [[0, 7], [7, 14], [14, 21]]
        belle_central = numpy.array([22.8, 56.5, 33.9]) * 1e-6
        belle_central = numpy.array([x / (b[1] - b[0]) for x, b in zip(belle_central, belle_bins)])
        belle_central = belle_central / (tauBplus / 2)

        belle_error = numpy.array([6.2, 10.6, 11.5]) * 1e-6
        belle_error = numpy.array([x / (b[1] - b[0]) for x, b in zip(belle_error, belle_bins)])
        belle_error = belle_error / (tauBplus / 2)

        belle_correlation = numpy.array([
            [100, 1.5, 10.3],
            [0, 100, 9.1],
            [0, 0, 100]
        ]) / 100.
        belle_correlation = (belle_correlation + belle_correlation.T) - numpy.diag(belle_correlation.diagonal())
        belle_covariance_statistical = numpy.diag(belle_error) @ belle_correlation @ numpy.diag(belle_error)

        systematic_error_belle_tracking = 0.7 * 0.01
        systematic_error_belle_pi0 = 2.0 * 0.01
        systematic_error_belle_pid = 1.0 * 0.01
        systematic_error_belle_kid = 2.0 * 0.01
        systematic_error_belle_qq = 0.0 * 0.01
        systematic_error_belle_crossfeed = 0.0 * 0.01
        systematic_error_belle_tagcal = 4.2 * 0.01
        systematic_error_belle_ffshape = 3.8 * 0.01

        systematic_error_belle_all = [
            systematic_error_belle_tracking,
            systematic_error_belle_pi0,
            systematic_error_belle_pid,
            systematic_error_belle_kid,
            systematic_error_belle_qq,
            systematic_error_belle_crossfeed,
            systematic_error_belle_tagcal,
            systematic_error_belle_ffshape,
        ]

        systematic_uncertainties_belle = [systematic * belle_central for systematic in systematic_error_belle_all]
        belle_covariance_systematic = sum([numpy.outer(s, s) for s in systematic_uncertainties_belle])

        belle_covariance = belle_covariance_systematic + belle_covariance_statistical

        self.bins = belle_bins
        self.central = belle_central
        self.covariance = belle_covariance
        self.covariance_statistical = belle_covariance_statistical
        self.covariance_systematic = belle_covariance_systematic
